(use-modules
  (guix packages)
  ((guix licenses) #:prefix license:)
  (guix download)
  (guix build-system gnu)
  (gnu packages)
  (gnu packages autotools)
  (gnu packages compression)
  (gnu packages guile)
  (gnu packages guile-xyz)
  (gnu packages pkg-config)
  (gnu packages texinfo))

(package
  (name "guile-lzlib")
  (version "0.2")
  (source "./guile-lzlib-0.2.tar.gz")
  (build-system gnu-build-system)
  (arguments `())
  (native-inputs
    `(("autoconf" ,autoconf)
      ("automake" ,automake)
      ("pkg-config" ,pkg-config)))
  (inputs `(("guile" ,guile-3.0)
            ("lzlib" ,lzlib)))
  (propagated-inputs `())
  (synopsis "")
  (description "")
  (home-page "")
  (license license:gpl3+))
